import asyncio
import webbrowser
import random
from urllib.parse import quote
from asyncio import Lock
from aiohttp import web
from aioalf.client import Client
from aioalf.manager import TokenManager


DEFAULT_PAGE = """<!DOCTYPE html><html><head><script>window.location = window.location.href.replace('#', '?');</script></head><body></body></html>"""  # noqa

CLOSE_PAGE = """<!DOCTYPE html><html><head></head><body><script>window.close();</script>
<h2 style="text-align: center;">
    Login successful.
    You can close this window.
</h2></body></html>
"""


class TokenStorage:

    def __init__(self):
        self.token = None

    def save(self, token):
        self.token = token

    def load(self):
        return self.token

    def clean(self):
        self.token = None


async def _run_web_server(port_range):
    async def token_handler(request):
        if 'access_token' not in request.query:
            return web.Response(body=DEFAULT_PAGE, content_type='text/html')

        token = {
            'access_token': request.query.get('access_token'),
            'token_type': request.query.get('token_type'),
            'expires_in': request.query.get('expires_in')
        }

        request.app['token_waiter'].set_result(token)

        return web.Response(body=CLOSE_PAGE, content_type='text/html')

    loop = asyncio.get_event_loop()

    app = web.Application()
    app['token_waiter'] = loop.create_future()
    app['port'] = random.randint(*port_range)

    app.router.add_get('/', token_handler)

    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, 'localhost', app['port'])
    await site.start()
    return app, site


class OAuthImplictTokenManager(TokenManager):

    app = None
    site = None
    browser_has_been_opened = False
    storage = None

    def __init__(self, token_endpoint,
                 client_id, client_secret, http_options=None,
                 scope=None):
        self._token_endpoint = token_endpoint
        self._client_id = client_id
        self._client_secret = client_secret
        self._scope = scope
        self._http_options = http_options if http_options else {}
        self._token_lock = Lock()

    def _shall_open_browser(cls):
        if not cls.browser_has_been_opened:
            cls.browser_has_been_opened = True
            return True
        else:
            return False

    async def get_token(self):
        async with self._token_lock:
            token = await self._retrieve_token()
            self.storage.save(token)
            return token

    def reset_token(self):
        self.storage.clean()

    def _get_stored_token(self):
        return self.storage.load()

    async def _retrieve_token(self):
        token = self._get_stored_token()
        if token:
            return token

        token = await self._ask_for_token()
        return token.get('access_token')

    async def _ask_for_token(self):
        token_url_template = (
            "{}/authorize?response_type=token&client_id={}&redirect_uri={}")

        token_url = token_url_template.format(
            self._token_endpoint,
            quote(self._client_id),
            quote("http://localhost:%d" % self.app['port']),
        )

        if self._scope:
            scope = self._scope
            if isinstance(scope, list):
                scope = " ".join(self._scope)
            token_url = "{}&scope={}".format(token_url, quote(scope))

        if self._shall_open_browser():
            webbrowser.open(token_url)

        result = None
        try:
            result = await self.app['token_waiter']
        except asyncio.CancelledError:
            await self.site.stop()
            return {'error': 'cancelled'}
        finally:
            await self.site.stop()
            return result


async def use_implicit_flow(token_storage, port_range=(32000, 32009)):
    app, site = await _run_web_server(port_range)
    OAuthImplictTokenManager.app = app
    OAuthImplictTokenManager.site = site
    OAuthImplictTokenManager.storage = token_storage
    Client.token_manager_class = OAuthImplictTokenManager



# -*- coding: utf-8 -*-
from base64 import b64encode
from aioalf.token import Token, TokenError, TokenHTTPError, TOKEN_FILTER
from aiohttp import ClientSession, ClientResponseError
from asyncio import Lock

import logging

logger = logging.getLogger(__name__)


class TokenManager(object):

    def __init__(self, token_endpoint, client_id,
                 client_secret, http_options=None,
                 scope=None):

        self._token_endpoint = token_endpoint
        self._client_id = client_id
        self._client_secret = client_secret
        self._scope = scope
        self._token = Token()
        self._http_options = http_options if http_options else {}
        self._token_lock = Lock()

    def _has_token(self):
        return self._token.is_valid()

    async def get_token(self):
        async with self._token_lock:
            if not self._has_token():
                await self._update_token()
            return self._token.access_token

    async def _get_token_data(self):
        return await self._request_token()

    def reset_token(self):
        self._token = Token()

    async def _update_token(self):
        token_data = await self._get_token_data()
        self._token = Token(token_data.get('access_token', ''),
                            token_data.get('expires_in', 0))

    async def _request_token(self):
        if not self._token_endpoint:
            raise TokenError('Missing token endpoint')

        data = {
            'grant_type': 'client_credentials',
        }

        if self._scope:
            scope = self._scope
            if isinstance(scope, list):
                scope = " ".join(self._scope)

            data['scope'] = scope

        return await self._fetch(
            url=self._token_endpoint,
            method="POST",
            auth=(self._client_id, self._client_secret),
            data=data
        )

    async def _fetch(self, url, method="GET", data=None, auth=None):
        request_data = dict(
            headers={},
            data=data
        )

        if auth is not None:
            try:
                passhash = b64encode(':'.join(auth).encode('ascii'))
            except TypeError as e:
                raise TokenError(
                    'Missing credentials (client_id:client_secret)', str(e)
                )

            request_data['headers']['Authorization'] = (
                'Basic %s' % passhash.decode('utf-8')
            )

        request_data.update(self._http_options)

        logger.debug('Request: %s %s', method, url)
        for header in request_data.get('headers'):
            if header.lower() == 'authorization':
                authorization_data = request_data.get('headers', {}).get(header)
                matches = TOKEN_FILTER.match(authorization_data)
                if matches:
                    logger.debug(('Header {}: {}<...>{}').format(header,
                                                                 matches.group('start'),
                                                                 matches.group('end')))
                continue
            logger.debug('Header %s: %s', header, request_data.get('headers', {}).get(header))

        try:
            async with ClientSession() as client:
                response = await client.request(method, url, **request_data)
                result = await response.json()
                return result
        except ClientResponseError as e:
            raise TokenHTTPError('Failed to request token', e.status, e.message)



#
# encoding: utf-8
import logging

from aiohttp import ClientSession
from aioalf.manager import TokenManager, TokenError
from aioalf.token import TOKEN_FILTER

BAD_TOKEN = 401

logger = logging.getLogger(__name__)


class Client(object):

    token_manager_class = TokenManager

    def __init__(self, client_id, client_secret,
                 token_endpoint, http_options=None,
                 scope=None):
        http_options = http_options is None and {} or http_options
        self._http_client = ClientSession()
        self._token_manager = self.token_manager_class(
            token_endpoint=token_endpoint,
            client_id=client_id,
            client_secret=client_secret,
            http_options=http_options,
            scope=scope)

    async def close(self):
        await self._http_client.close()

    async def request(self, method, url, **kwargs):
        try:
            response = await self._authorized_fetch(method,
                                                    url,
                                                    **kwargs)
            if response.status != BAD_TOKEN:
                return response

            self._token_manager.reset_token()
            response = await self._authorized_fetch(method,
                                                    url,
                                                    **kwargs)
            return response

        except TokenError:
            self._token_manager.reset_token()
            raise

    async def _authorized_fetch(self, method, url, **kwargs):
        access_token = await self._token_manager.get_token()

        auth_headers = {'Authorization': 'Bearer {}'.format(access_token)}
        if 'headers' in kwargs:
            kwargs['headers'].update(auth_headers)
        else:
            kwargs['headers'] = auth_headers

        logger.debug('Request: %s %s', method, url)
        for header in kwargs.get('headers'):
            if header.lower() == 'authorization':
                authorization_data = kwargs.get('headers', {}).get(header)
                matches = TOKEN_FILTER.match(authorization_data)
                if matches:
                    logger.debug(('Header {}: {}<...>{}').format(header,
                                                                 matches.group('start'),
                                                                 matches.group('end')))
                continue
            logger.debug('Header %s: %s', header, kwargs.get('headers').get(header))

        return await self._http_client.request(method, url, **kwargs)

    def __enter__(self):
        raise TypeError("Use async with instead.")

    def __exit__(self, type, value, traceback):
        pass  # pragma: no cover

    async def __aenter__(self):
        return self

    async def __aexit__(self, type, value, traceback):
        await self.close()



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from aioalf import __version__

tests_require = [
    'mock',
    'asynctest',
    'bumpversion',
    'nose',
    'coverage',
    'yanc',
    'ipdb',
    'tox',
    'flake8',
]

setup(
    name='aio-alf',
    version=__version__,
    description="OAuth Client For aiohttp",
    long_description=open('README.rst').read(),
    keywords='oauth client client_credentials aiohttp',
    author=u'Globo.com',
    author_email='entretenimento1@corp.globo.com',
    url='https://github.com/globocom/aio-alf',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(
        exclude=(
            'tests',
        ),
    ),
    include_package_data=True,
    install_requires=[
        'aiohttp>=2.3.0',
    ],
    extras_require={
        'tests': tests_require,
    },
)
